#include <stdlib.h>
#include <string.h>

#ifdef ENABLE_NLS
#include <libintl.h>
#endif

#define         _ISOC9X_SOURCE  1
#define         _ISOC99_SOURCE  1
#define         __USE_ISOC99    1
#define         __USE_ISOC9X    1

#include <math.h>
#include "utils.h"
#include "ladspa.h"

#line 29 "foo_transients_mono.xml"

// define gain scale to scale the parameterized gain to better suit average envelopes
#define FOO_ARCHITECT_GAIN_SCALE 5.0f

#define FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE 0
#define FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB 1
#define FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB 2
#define FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN 3
#define FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN 4
#define FOO_TRANSIENTS_MONO_LATENCY    5
#define FOO_TRANSIENTS_MONO_INPUT      6
#define FOO_TRANSIENTS_MONO_OUTPUT     7

static LADSPA_Descriptor *foo_transients_monoDescriptor = NULL;

typedef struct {
	LADSPA_Data *average_difference;
	LADSPA_Data *attack_gain_db;
	LADSPA_Data *release_gain_db;
	LADSPA_Data *actual_attack_gain;
	LADSPA_Data *actual_release_gain;
	LADSPA_Data *latency;
	LADSPA_Data *input;
	LADSPA_Data *output;
	rms *        fast_rms;
	rms *        slow_rms;
	LADSPA_Data run_adding_gain;
} Foo_transients_mono;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {

	switch (index) {
	case 0:
		return foo_transients_monoDescriptor;
	default:
		return NULL;
	}
}

static void cleanupFoo_transients_mono(LADSPA_Handle instance) {
#line 44 "foo_transients_mono.xml"
	Foo_transients_mono *plugin_data = (Foo_transients_mono *)instance;
	rms_free (plugin_data->fast_rms);
	rms_free (plugin_data->slow_rms);
	free(instance);
}

static void connectPortFoo_transients_mono(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	Foo_transients_mono *plugin;

	plugin = (Foo_transients_mono *)instance;
	switch (port) {
	case FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE:
		plugin->average_difference = data;
		break;
	case FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB:
		plugin->attack_gain_db = data;
		break;
	case FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB:
		plugin->release_gain_db = data;
		break;
	case FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN:
		plugin->actual_attack_gain = data;
		break;
	case FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN:
		plugin->actual_release_gain = data;
		break;
	case FOO_TRANSIENTS_MONO_LATENCY:
		plugin->latency = data;
		break;
	case FOO_TRANSIENTS_MONO_INPUT:
		plugin->input = data;
		break;
	case FOO_TRANSIENTS_MONO_OUTPUT:
		plugin->output = data;
		break;
	}
}

static LADSPA_Handle instantiateFoo_transients_mono(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	Foo_transients_mono *plugin_data = (Foo_transients_mono *)malloc(sizeof(Foo_transients_mono));
	rms *fast_rms = NULL;
	rms *slow_rms = NULL;

#line 39 "foo_transients_mono.xml"
	slow_rms = rms_new((float)s_rate, 0.06f);
	fast_rms = rms_new((float)s_rate, 0.01f);

	plugin_data->fast_rms = fast_rms;
	plugin_data->slow_rms = slow_rms;

	return (LADSPA_Handle)plugin_data;
}

#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b = v)
#define RUN_ADDING    0
#define RUN_REPLACING 1

static void runFoo_transients_mono(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_transients_mono *plugin_data = (Foo_transients_mono *)instance;

	/* Averaging difference (s) (float value) */
	const LADSPA_Data average_difference = *(plugin_data->average_difference);

	/* Attack gain (dB) (float value) */
	const LADSPA_Data attack_gain_db = *(plugin_data->attack_gain_db);

	/* Release gain (dB) (float value) */
	const LADSPA_Data release_gain_db = *(plugin_data->release_gain_db);

	/* Input L (array of floats of length sample_count) */
	const LADSPA_Data * const input = plugin_data->input;

	/* Output L (array of floats of length sample_count) */
	LADSPA_Data * const output = plugin_data->output;
	rms * fast_rms = plugin_data->fast_rms;
	rms * slow_rms = plugin_data->slow_rms;

#line 49 "foo_transients_mono.xml"
	// DSP
	unsigned long i;
	float tmp, fast, slow;

	float relevant_attack_gain  = 0.0f;
	float relevant_release_gain = 0.0f;

#define BASELINE 0.05f
	rms_set_time(slow_rms, BASELINE + average_difference / 2);
	rms_set_time(fast_rms, BASELINE - average_difference / 2);
#undef BASELINE

	for (i = 0; i < sample_count; i++) {

	    slow = rms_run(slow_rms, input[i]);
	    fast = rms_run(fast_rms, input[i]);

	    if (fast > slow) {
	        // Attack
	        tmp = (fast - slow) * attack_gain_db * FOO_ARCHITECT_GAIN_SCALE;
	
	        if ( ( attack_gain_db < 0.0f && tmp < relevant_attack_gain) ||
	             ( attack_gain_db > 0.0f && tmp > relevant_attack_gain) )
	                relevant_attack_gain = tmp;
	                
	    } else {
	        // Release
	        tmp = (slow - fast) * release_gain_db * FOO_ARCHITECT_GAIN_SCALE;

	        if ( ( release_gain_db < 0.0f && tmp < relevant_release_gain) ||
	             ( release_gain_db > 0.0f && tmp > relevant_release_gain) )
	                relevant_release_gain = tmp;

	        //printf("Release gain: %f, envelope difference = %f\n", tmp, slow - fast);
	    }

	    buffer_write(output[i],  input[i]  * DB_CO(tmp));

#ifdef FOO_TESTER
	    //buffer_write(output[i], fast);
#endif 
	}

	*(plugin_data->actual_attack_gain)  = relevant_attack_gain;
	*(plugin_data->actual_release_gain) = relevant_release_gain;
}
#undef buffer_write
#undef RUN_ADDING
#undef RUN_REPLACING

#define buffer_write(b, v) (b += (v) * run_adding_gain)
#define RUN_ADDING    1
#define RUN_REPLACING 0

static void setRunAddingGainFoo_transients_mono(LADSPA_Handle instance, LADSPA_Data gain) {
	((Foo_transients_mono *)instance)->run_adding_gain = gain;
}

static void runAddingFoo_transients_mono(LADSPA_Handle instance, unsigned long sample_count) {
	Foo_transients_mono *plugin_data = (Foo_transients_mono *)instance;
	LADSPA_Data run_adding_gain = plugin_data->run_adding_gain;

	/* Averaging difference (s) (float value) */
	const LADSPA_Data average_difference = *(plugin_data->average_difference);

	/* Attack gain (dB) (float value) */
	const LADSPA_Data attack_gain_db = *(plugin_data->attack_gain_db);

	/* Release gain (dB) (float value) */
	const LADSPA_Data release_gain_db = *(plugin_data->release_gain_db);

	/* Input L (array of floats of length sample_count) */
	const LADSPA_Data * const input = plugin_data->input;

	/* Output L (array of floats of length sample_count) */
	LADSPA_Data * const output = plugin_data->output;
	rms * fast_rms = plugin_data->fast_rms;
	rms * slow_rms = plugin_data->slow_rms;

#line 49 "foo_transients_mono.xml"
	// DSP
	unsigned long i;
	float tmp, fast, slow;

	float relevant_attack_gain  = 0.0f;
	float relevant_release_gain = 0.0f;

#define BASELINE 0.05f
	rms_set_time(slow_rms, BASELINE + average_difference / 2);
	rms_set_time(fast_rms, BASELINE - average_difference / 2);
#undef BASELINE

	for (i = 0; i < sample_count; i++) {

	    slow = rms_run(slow_rms, input[i]);
	    fast = rms_run(fast_rms, input[i]);

	    if (fast > slow) {
	        // Attack
	        tmp = (fast - slow) * attack_gain_db * FOO_ARCHITECT_GAIN_SCALE;
	
	        if ( ( attack_gain_db < 0.0f && tmp < relevant_attack_gain) ||
	             ( attack_gain_db > 0.0f && tmp > relevant_attack_gain) )
	                relevant_attack_gain = tmp;
	                
	    } else {
	        // Release
	        tmp = (slow - fast) * release_gain_db * FOO_ARCHITECT_GAIN_SCALE;

	        if ( ( release_gain_db < 0.0f && tmp < relevant_release_gain) ||
	             ( release_gain_db > 0.0f && tmp > relevant_release_gain) )
	                relevant_release_gain = tmp;

	        //printf("Release gain: %f, envelope difference = %f\n", tmp, slow - fast);
	    }

	    buffer_write(output[i],  input[i]  * DB_CO(tmp));

#ifdef FOO_TESTER
	    //buffer_write(output[i], fast);
#endif 
	}

	*(plugin_data->actual_attack_gain)  = relevant_attack_gain;
	*(plugin_data->actual_release_gain) = relevant_release_gain;
}

#ifndef FOO_TESTER
void _init() {
#else
void fake_init() {
#endif
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

#ifdef ENABLE_NLS
#define D_(s) dgettext(PACKAGE, s)
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, PACKAGE_LOCALE_DIR);
#else
#define D_(s) (s)
#endif


	foo_transients_monoDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (foo_transients_monoDescriptor) {
		foo_transients_monoDescriptor->UniqueID = 3183;
		foo_transients_monoDescriptor->Label = "foo_transients_mono";
		foo_transients_monoDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		foo_transients_monoDescriptor->Name =
		 D_("Foo Transient Architect (mono)");
		foo_transients_monoDescriptor->Maker =
		 "Sampo Savolainen <v2@iki.fi>";
		foo_transients_monoDescriptor->Copyright =
		 "GPL";
		foo_transients_monoDescriptor->PortCount = 8;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(8,
		 sizeof(LADSPA_PortDescriptor));
		foo_transients_monoDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(8,
		 sizeof(LADSPA_PortRangeHint));
		foo_transients_monoDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(8, sizeof(char*));
		foo_transients_monoDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Averaging difference (s) */
		port_descriptors[FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE] =
		 D_("Averaging difference (s)");
		port_range_hints[FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_HIGH;
		port_range_hints[FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE].LowerBound = 0.01;
		port_range_hints[FOO_TRANSIENTS_MONO_AVERAGE_DIFFERENCE].UpperBound = 0.1;

		/* Parameters for Attack gain (dB) */
		port_descriptors[FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB] =
		 D_("Attack gain (dB)");
		port_range_hints[FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0;
		port_range_hints[FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB].LowerBound = -30.0;
		port_range_hints[FOO_TRANSIENTS_MONO_ATTACK_GAIN_DB].UpperBound = +30.0;

		/* Parameters for Release gain (dB) */
		port_descriptors[FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB] =
		 D_("Release gain (dB)");
		port_range_hints[FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_0;
		port_range_hints[FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB].LowerBound = -30.0;
		port_range_hints[FOO_TRANSIENTS_MONO_RELEASE_GAIN_DB].UpperBound = +30.0;

		/* Parameters for Current attack gain (dB) */
		port_descriptors[FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN] =
		 D_("Current attack gain (dB)");
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN].LowerBound = -30.0;
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_ATTACK_GAIN].UpperBound = +30.0;

		/* Parameters for Current release gain (dB) */
		port_descriptors[FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN] =
		 D_("Current release gain (dB)");
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN].LowerBound = -30.0;
		port_range_hints[FOO_TRANSIENTS_MONO_ACTUAL_RELEASE_GAIN].UpperBound = +30.0;

		/* Parameters for latency */
		port_descriptors[FOO_TRANSIENTS_MONO_LATENCY] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_CONTROL;
		port_names[FOO_TRANSIENTS_MONO_LATENCY] =
		 D_("latency");
		port_range_hints[FOO_TRANSIENTS_MONO_LATENCY].HintDescriptor = 0;

		/* Parameters for Input L */
		port_descriptors[FOO_TRANSIENTS_MONO_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_TRANSIENTS_MONO_INPUT] =
		 D_("Input L");
		port_range_hints[FOO_TRANSIENTS_MONO_INPUT].HintDescriptor = 0;

		/* Parameters for Output L */
		port_descriptors[FOO_TRANSIENTS_MONO_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[FOO_TRANSIENTS_MONO_OUTPUT] =
		 D_("Output L");
		port_range_hints[FOO_TRANSIENTS_MONO_OUTPUT].HintDescriptor = 0;

		foo_transients_monoDescriptor->activate = NULL;
		foo_transients_monoDescriptor->cleanup = cleanupFoo_transients_mono;
		foo_transients_monoDescriptor->connect_port = connectPortFoo_transients_mono;
		foo_transients_monoDescriptor->deactivate = NULL;
		foo_transients_monoDescriptor->instantiate = instantiateFoo_transients_mono;
		foo_transients_monoDescriptor->run = runFoo_transients_mono;
		foo_transients_monoDescriptor->run_adding = runAddingFoo_transients_mono;
		foo_transients_monoDescriptor->set_run_adding_gain = setRunAddingGainFoo_transients_mono;
	}
}

#ifndef FOO_TESTER
void _fini() {
#else
void fake_fini() {
#endif
	if (foo_transients_monoDescriptor) {
		free((LADSPA_PortDescriptor *)foo_transients_monoDescriptor->PortDescriptors);
		free((char **)foo_transients_monoDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)foo_transients_monoDescriptor->PortRangeHints);
		free(foo_transients_monoDescriptor);
	}

}
